import numpy as np
from .base_signal import BaseSignal

__all__ = ['CAR']


class CAR(BaseSignal):
    """Signal generatpr for continuously autoregressive (CAR) signals.

    Parameters
    ----------
    ar_param : number (default 1.0)
        Parameter of the AR(1) process
    sigma : number (default 1.0)
        Standard deviation of the signal
    start_value : number (default 0.0)
        Starting value of the AR process
        
    """

    def __init__(self, ar_param=1.0, sigma=0.5, start_value=0.01):
        self.vectorizable = False
        self.ar_param = ar_param
        self.sigma = sigma
        self.start_value = start_value
        self.previous_value = None
        self.previous_time = None

    def sample_next(self, time, samples, errors):
        """Sample a single time point

        Parameters
        ----------
        time : number
            Time at which a sample was required

        Returns
        -------
        float
            sampled signal for time t

        """
        if self.previous_value is None:
            output = self.start_value
        else:
            time_diff = time - self.previous_time
            noise = np.random.normal(loc=0.0, scale=1.0, size=1)
            output = (np.power(self.ar_param, time_diff))*self.previous_value+\
                self.sigma*np.sqrt(1-np.power(self.ar_param, time_diff))*noise
        self.previous_time = time
        self.previous_value = output
        return output



import warnings
import numpy as np
from .base_signal import BaseSignal
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from jitcdde import y, t, jitcdde


__all__ = ['MackeyGlass']


class MackeyGlass(BaseSignal):
    """Signal generator for the Mackey-Glass delay differential equation (DDE).

    The Mackey-Glass DDE is defined as follows:
    .. math::

        \\frac{dx}{dt} = \\beta \\frac{ x_{\\tau} }{1+{x_{\\tau}}^n}-\\gamma x,
        \\quad \\gamma,\\beta,n > 0,
        x_{\\tau} = x(t - \\tau)

    Chaotic behavior may occur for tau >= 17.

    Parameters
    ----------
    tau : float (default 17)
        The delay parameter
    n : float (default 10)
        The parameter n
    beta : float (default 0.2)
        The parameter beta
    gamma : float (default 0.1)
        The parameter gamma
    initial_condition : array-like or None (default None)
        A 2-D array consisting of entries in the form (time, value, derivative)
        to be used as an inital condition on the DDE. If set to None a default
        will be used.
    burn_in : float (default 500)
        Amount of time after which samples will be taken and returned

    """

    def __init__(self, tau=17., n=10., beta=0.2, gamma=0.1, initial_condition=None, burn_in=500):
        self.vectorizable = True

        # Set system of equations
        f = [- gamma * y(0) + beta * y(0, t-tau) / (1.0 + y(0, t-tau) ** n)]
        self.dde = jitcdde(f)

        # Set initial condition
        if initial_condition is None:
            y_initial = 0.5
            dy = lambda y: -gamma * y + beta * y / (1.0 + y ** n)
            dy_initial = dy(y_initial)
            self.dde.add_past_point(0.0, np.array([y_initial]), np.array([dy_initial]))
            self.dde.add_past_point(tau, np.array([1.0]), np.array([0.0]))
        else:
            for condition in initial_condition:
                time, value, derivative = condition
                self.dde.add_past_point(time, np.array([value]), np.array([derivative]))

        # Prepare DDE
        self.dde.generate_lambdas()
        self.dde.set_integration_parameters()

        # Run burn_in
        self.burn_in = burn_in
        self.dde.integrate_blindly(self.burn_in)

    def sample_next(self, time, samples, errors):
        """Samples next point based on history of samples and errors

        Parameters
        ----------
        time : int
            time
        samples : array-like
            all samples taken so far
        errors : array-like
            all errors sampled so far

        Returns
        -------
        float
            sampled signal for time t

        """
        return self.dde.integrate(self.burn_in + time)

    def sample_vectorized(self, time_vector):
        """Samples for all time points in input

        Parameters
        ----------
        time_vector : array like
            all time stamps to be sampled

        Returns
        -------
        numpy array
            samples for times provided in time_vector

        """
        samples = []
        for t in time_vector:
            samples.append(self.dde.integrate(self.burn_in + t))
        return np.array(samples).reshape(-1,)



import numpy as np


__all__ = ['TimeSampler']


class TimeSampler:
    """TimeSampler determines how and when samples will be taken from signal and noise.
    Samples timestamps for regular and irregular time signals

    Parameters
    ----------
    start_time: float/int (default 0)
                Time sampling of time series starts
    stop_time: float/int (default 10)
                Time sampling of time series stops

    """
    def __init__(self, start_time=0, stop_time=10):
        self.start_time = start_time
        self.stop_time = stop_time

    def sample_regular_time(self, num_points=None, resolution=None):
        """
        Samples regularly spaced time using the number of points or the
        resolution of the signal. Only one of the parameters is to be
        initialized. The resolution keyword argument is given priority.

        Parameters
        ----------
        num_points: int (default None)
            Number of points in time series
        resolution: float/int (default None)
            Resolution of the time series

        Returns
        -------
        numpy array
            Regularly sampled timestamps

        """
        if num_points is None and resolution is None:
            raise ValueError("One of the keyword arguments must be initialized.")
        if resolution is not None:
            time_vector = np.arange(self.start_time, self.stop_time,
                                    resolution)
            return time_vector
        else:
            time_vector = np.linspace(self.start_time, self.stop_time,
                                      num_points)
            return time_vector

    def sample_irregular_time(self, num_points=None, resolution=None,
                              keep_percentage=100):
        """
        Samples regularly spaced time using the number of points or the
        resolution of the signal. Only one of the parameters is to be
        initialized. The resolution keyword argument is given priority.

        Parameters
        ----------
        num_points: int (default None)
            Number of points in time series
        resolution: float/int (default None)
            Resolution of the time series
        keep_percentage: int(default 100)
            Percentage of points to be retained in the irregular series

        Returns
        -------
        numpy array
            Irregularly sampled timestamps

        """
        if num_points is None and resolution is None:
            raise ValueError("One of the keyword arguments must be initialized.")
        if resolution is not None:
            time_vector = np.arange(self.start_time, self.stop_time,
                                    resolution)
        else:
            time_vector = np.linspace(self.start_time, self.stop_time,
                                      num_points)
            resolution = float(self.stop_time-self.start_time)/num_points
        time_vector = self._select_random_indices(time_vector,
                                                  keep_percentage)
        return self._create_perturbations(time_vector, resolution)

    def _create_perturbations(self, time_vector, resolution):
        """
        Internal functions to create perturbations in timestamps

        Parameters
        ----------
        time_vector: numpy array
            timestamp vector

        resolution: float/int
            resolution of the time series

        Returns
        -------
        numpy array
            Irregularly sampled timestamps with perturbations

        """
        sample_perturbations = np.random.normal(loc=0.0, scale=resolution,
                                                size=len(time_vector))
        time_vector = time_vector + sample_perturbations
        return np.sort(time_vector)

    def _select_random_indices(self, time_vector, keep_percentage):
        """
        Internal functions to randomly select timestamps

        Parameters
        ----------
        time_vector: numpy array
            timestamp vector

        keep_percentage: float/int
            percentage of points retained

        Returns
        -------
        numpy array
            Irregularly sampled timestamps

        """
        num_points = len(time_vector)
        num_select_points = int(keep_percentage*num_points/100)
        index = np.sort(np.random.choice(num_points, size=num_select_points,
                        replace=False))
        return time_vector[index]



import numpy as np
from .base_noise import BaseNoise


__all__ = ['RedNoise']


class RedNoise(BaseNoise):
    """Red noise generator.
    This class adds correlated (red) noise to your signal.

    Attributes
    ----------
    mean : float
        mean for the noise
    std : float
        standard deviation for the noise
    tau : float
        ?
    start_value : float
        ?

    """

    def __init__(self, mean=0, std=1., tau=0.2, start_value=0):
        self.vectorizable = False
        self.mean = mean
        self.std = std
        self.start_value = 0
        self.tau = tau
        self.previous_value = None
        self.previous_time = None

    def sample_next(self, t, samples, errors):
        if self.previous_time is None:
            red_noise = self.start_value
        else:
            time_diff = t - self.previous_time
            wnoise = np.random.normal(loc=self.mean, scale=self.std, size=1)
            red_noise = ((self.tau/(self.tau + time_diff)) *
                         (time_diff*wnoise + self.previous_value))
        self.previous_time = t
        self.previous_value =red_noise
        return red_noise



import numpy as np
from .base_noise import BaseNoise


__all__ = ['GaussianNoise']


class GaussianNoise(BaseNoise):
    """Gaussian noise generator.
    This class adds uncorrelated, additive white noise to your signal.

    Attributes
    ----------
    mean : float
        mean for the noise
    std : float
        standard deviation for the noise

    """

    def __init__(self, mean=0, std=1.):
        self.vectorizable = True
        self.mean = mean
        self.std = std

    def sample_next(self, t, samples, errors):
        return np.random.normal(loc=self.mean, scale=self.std, size=1)

    def sample_vectorized(self, time_vector):
        n_samples = len(time_vector)
        return np.random.normal(loc=self.mean, scale=self.std, size=n_samples)



import numpy as np
import scipy.special
from .base_signal import BaseSignal

__all__ = ['GaussianProcess']


class GaussianProcess(BaseSignal):
    """Gaussian Process time series sampler
    
    Samples time series from Gaussian Process with selected covariance function (kernel).

    Parameters
    ----------
    kernel : {'SE', 'Constant', 'Exponential', 'RQ', 'Linear', 'Matern', 'Periodic'}
        the kernel type, as described in [1]_ and [2]_, which can be:
        
        - `Constant`. All covariances set to `variance`
        - `Exponential`. Ornstein-Uhlenbeck kernel. Optionally, set keyword `gamma` for a gamma-exponential kernel
        - `SE`, the squared exponential.
        - `RQ`, the rational quadratic. To use this kernel, set keyword argument `alpha`
        - `Linear`. To use this kernel, set keyword arguments `c` and `offset`
        - `Matern`. To use this kernel, set keyword argument `nu`
        - `Periodic`. To use this kernel, set keyword argument `p` for the period
        
    mean : float
        the mean of the gaussian process
    variance : float
        the output variance of the gaussian process (sigma^2)
    lengthscale : float
            the characteristic lengthscale used to generate the covariance matrix
    
    References
    ----------
    .. [1] URL: http://www.cs.toronto.edu/~duvenaud/cookbook/index.html
    .. [2] Rasmussen, C.E., 2006. Gaussian processes for machine learning. URL: https://pdfs.semanticscholar.org/a9fe/ab0fe858dbde2eecff8b1f7c629cc6aff8ad.pdf

    """

    def __init__(self, kernel="SE", lengthscale=1., mean=0., variance=1., c=1., gamma=1., alpha=1., offset=0., nu=5./2, p=1.):
        self.vectorizable = True
        self.lengthscale = lengthscale
        self.mean = mean
        self.variance = variance
        self.kernel = kernel
        self.kernel_function = {"Constant": lambda x1, x2: variance,
                                "Exponential": lambda x1, x2: variance * np.exp(-np.power(np.abs(x1 - x2) / lengthscale, gamma)),
                                "SE": lambda x1, x2: variance * np.exp(- np.square(x1 - x2) / (2 * np.square(lengthscale))),
                                "RQ": lambda x1, x2: variance * np.power((1 + np.square(x1 - x2) / (2 * alpha * np.square(lengthscale))), -alpha),
                                "Linear": lambda x1, x2: variance * (x1 - c) * (x2 - c) + offset,
                                "Matern": lambda x1, x2: variance if x1 - x2 == 0. else variance * (np.power(2, 1 - nu) / scipy.special.gamma(nu)) * np.power(np.sqrt(2 * nu) * np.abs(x1 - x2) / lengthscale, nu) * scipy.special.kv(nu, np.sqrt(2 * nu) * np.abs(x1 - x2) / lengthscale),
                                "Periodic":lambda x1, x2: variance * np.exp(- 2 * np.square(np.sin(np.pi * np.abs(x1 - x2) / p))),
                                }[kernel]

    def sample_next(self, time, samples, errors):
        """Sample a single time point

        Parameters
        ----------
        time : number
            Time at which a sample was required

        Returns
        -------
        float
            sampled signal for time t

        """
        raise NotImplementedError

    def sample_vectorized(self, time_vector):
        """Sample entire series based off of time vector

        Parameters
        ----------
        time_vector : array-like
            Timestamps for signal generation

        Returns
        -------
        array-like
            sampled signal for time vector

        """
        cartesian_time = np.dstack(np.meshgrid(time_vector, time_vector)).reshape(-1, 2)
        covariance_matrix = (np.vectorize(self.kernel_function)(cartesian_time[:, 0], cartesian_time[:, 1])).reshape(-1, time_vector.shape[0])
        covariance_matrix[np.diag_indices_from(covariance_matrix)] += 1e-12  # Add small value to diagonal for numerical stability
        return np.random.multivariate_normal(mean=np.full(shape=(time_vector.shape[0],), fill_value=self.mean), cov=covariance_matrix)



import numpy as np
from .base_signal import BaseSignal


__all__ = ['Sinusoidal']


class Sinusoidal(BaseSignal):
    """Signal generator for harmonic (sinusoidal) waves.

    Parameters
    ----------
    amplitude : number (default 1.0)
        Amplitude of the harmonic series
    frequency : number (default 1.0)
        Frequency of the harmonic series
    ftype : function (default np.sin)
        Harmonic function

    """

    def __init__(self, amplitude=1.0, frequency=1.0, ftype=np.sin):
        self.vectorizable = True
        self.amplitude = amplitude
        self.ftype = ftype
        self.frequency = frequency

    def sample_next(self, time, samples, errors):
        """Sample a single time point

        Parameters
        ----------
        time : number
            Time at which a sample was required

        Returns
        -------
        float
            sampled signal for time t

        """
        return self.amplitude * self.ftype(2*np.pi*self.frequency*time)

    def sample_vectorized(self, time_vector):
        """Sample entire series based off of time vector

        Parameters
        ----------
        time_vector : array-like
            Timestamps for signal generation

        Returns
        -------
        array-like
            sampled signal for time vector

        """
        if self.vectorizable is True:
            signal = self.amplitude * self.ftype(2*np.pi*self.frequency *
                                                 time_vector)
            return signal
        else:
            raise ValueError("Signal type not vectorizable")



from os.path import dirname as _dirname, basename as _basename, isfile as _isfile
import glob as _glob

exec('\n'.join(map(lambda name: "from ." + name + " import *", 
                   [_basename(f)[:-3] for f in _glob.glob(_dirname(__file__) + "/*.py") \
                    if _isfile(f) and not _basename(f).startswith('_')])))



import numpy as np
from .base_signal import BaseSignal


__all__ = ['NARMA']


class NARMA(BaseSignal):
    """Non-linear Autoregressive Moving Average generator.
    
    An n-th order signal generator of the NARMA class, as defined in [1]_:
    .. \math:: y(k+1) = a[0] * y(k) + a[1] * y(k) * \sum_{i=0}^{n-1} y(k-i) + a[2] * u(k-(n-1)) * u(k) + a[3]
    
    where u is generated from Uniform(0, 0.5).
    
    NOTE: Only supports regular time samples.
    
    Parameters
    ----------
    order : int (default 10)
        The order (n) of non-linear interactions as described in the formula above.
    coefficients : iterable (default [0.3, 0.05, 1.5, 0.1])
        The coefficients denoted by iterable `a` in the formula above. As in [1]_.
    initial_condition : iterable or None (default None)
        An array of starting values of y(k-n) until y(k). The default is an aray of zeros.
    seed : int
        Use this seed to recreate any of the internal errors.
        
    Attributes
    ----------
    errors : numpy array or None
        Random number sequence that was used to generate last NARMA sequence.
    
    References
    ----------
    .. [1] http://ieeexplore.ieee.org.ezp-prod1.hul.harvard.edu/stamp/stamp.jsp?arnumber=846741
    
    """
    
    def __init__(self, order=10, coefficients=[0.3, 0.05, 1.5, 0.1], initial_condition=None,
                 error_initial_condition=None, seed=42):
        self.vectorizable = True
        self.order = order
        self.coefficients = np.array(coefficients)
        self.random = np.random.RandomState(seed)
        
        # Store initial conditions
        if initial_condition is None:
            self.initial_condition = np.zeros(order)
        else:
            self.initial_condition = np.array(initial_condition)
        
        # You may provide an error initial condition
        if error_initial_condition is None:
            self.error_initial_condition = self.random.uniform(0, 0.5, size=order)
        else:
            self.error_initial_condition = error_initial_condition
        
    def _next_value(self, values, rands, index):
        """Internal short-hand method to calculate next value."""
        # Short-hand parameters
        n = self.order
        a = self.coefficients
        
        # Get value arrays
        i = index
        y = values
        u = rands
        
        # Compute next value
        return a[0] * y[i-1] + a[1] * y[i-1] * np.sum(y[i-n:n]) + a[2] * u[i-n] * u[i] + a[3]
        
    def sample_next(self, time, samples, errors):
        """This method is not available for NARMA, due to internal error sampling."""
        raise NotImplementedError("NARMA can only be sampled vectorized.")
        
    def sample_vectorized(self, times):
        """Samples for all time points in input
        
        Internalizes Uniform(0, 0.5) random distortion for u.

        Parameters
        ----------
        times: array like
            all time stamps to be sampled
        
        Returns
        -------
        samples : numpy array
            samples for times provided in time_vector

        """
        # Set bounds
        start = self.initial_condition.shape[0]
        
        # Get relevant arrays
        inits = self.initial_condition
        rand_inits = self.error_initial_condition
        rands = np.concatenate((rand_inits, self.random.uniform(0, .5, size=times.shape[0])))
        values = np.concatenate((inits, np.zeros(times.shape[0])))
        
        # Sample step-wise
        end = values.shape[0]
        for t in range(start, end):
            values[t] = self._next_value(values, rands, t)
        
        # Store valus for later retrieval
        self.errors = rands[start:]
        
        # Return trimmed values (exclude initial condition)
        samples = values[start:]
        return samples



import numpy as np
from .base_signal import BaseSignal

__all__ = ['PseudoPeriodic']


class PseudoPeriodic(BaseSignal):
    """Signal generator for pseudoeriodic waves. 
    
    The wave's amplitude and frequency have some stochasticity that 
    can be set manually.

    Parameters
    ----------
    amplitude : number (default 1.0)
        Amplitude of the harmonic series
    frequency : number (default 1.0)
        Frequency of the harmonic series
    ampSD : number (default 0.1)
        Amplitude standard deviation
    freqSD : number (default 0.1)
        Frequency standard deviation
    ftype : function(default np.sin)
        Harmonic function
        
    """
    
    def __init__(self, amplitude=1.0, frequency=100, ampSD=0.1, freqSD=0.4,
                 ftype=np.sin):
        self.vectorizable = True
        self.amplitude = amplitude
        self.frequency = frequency
        self.freqSD = freqSD
        self.ampSD = ampSD
        self.ftype = ftype

    def sample_next(self, time, samples, errors):
        """Sample a single time point

        Parameters
        ----------
        time : number
            Time at which a sample was required

        Returns
        -------
        float
            sampled signal for time t

        """
        freq_val = np.random.normal(loc=self.frequency, scale=self.freqSD, size=1)
        amplitude_val = np.random.normal(loc=self.amplitude, scale=self.ampSD, size=1)
        return float(amplitude_val * np.sin(freq_val * time))

    def sample_vectorized(self, time_vector):
        """Sample entire series based off of time vector

        Parameters
        ----------
        time_vector : array-like
            Timestamps for signal generation

        Returns
        -------
        array-like
            sampled signal for time vector

        """
        n_samples = len(time_vector)
        freq_arr = np.random.normal(loc=self.frequency, scale=self.freqSD, size=n_samples)
        amp_arr = np.random.normal(loc=self.amplitude, scale=self.ampSD, size=n_samples)
        signal = np.multiply(amp_arr, self.ftype(np.multiply(freq_arr, time_vector)))
        return signal



import numpy as np

__all__ = ['TimeSeries']


class TimeSeries:
    """A TimeSeries object is the main interface from which to sample time series.
    You have to provide at least a signal generator; a noise generator is optional.
    It is recommended to set the sampling frequency.

    Parameters
    ----------
    signal_generator : Signal object
        signal object for time series
    noise_generator : Noise object
        noise object for time series

    """
    def __init__(self, signal_generator, noise_generator=None):
        self.signal_generator = signal_generator
        self.noise_generator = noise_generator


    def sample(self, time_vector):
        """Samples from the specified TimeSeries.
        
        Parameters
        ----------
        time_vector : numpy array
            Times at which to generate a sample
            
        Returns
        -------
        samples, signals, errors, : tuple (array, array, array)
            Returns samples, and the signals and errors they were constructed from
        """
        
        # Vectorize if possible
        if self.signal_generator.vectorizable and not self.noise_generator is None and self.noise_generator.vectorizable:
            signals = self.signal_generator.sample_vectorized(time_vector)
            errors = self.noise_generator.sample_vectorized(time_vector)
            samples = signals + errors
        elif self.signal_generator.vectorizable and self.noise_generator is None:
            signals = self.signal_generator.sample_vectorized(time_vector)
            errors = np.zeros(len(time_vector))
            samples = signals
        else:
            n_samples = len(time_vector)
            samples = np.zeros(n_samples)  # Signal and errors combined
            signals = np.zeros(n_samples)  # Signal samples
            errors = np.zeros(n_samples)  # Handle errors seprately
            times = np.arange(n_samples)

            # Sample iteratively, while providing access to all previously sampled steps
            for i in range(n_samples):
                # Get time
                t = time_vector[i]
                # Sample error
                if not self.noise_generator is None:
                    errors[i] = self.noise_generator.sample_next(t, samples[:i - 1], errors[:i - 1])

                # Sample signal
                signal = self.signal_generator.sample_next(t, samples[:i - 1], errors[:i - 1])
                signals[i] = signal

                # Compound signal and noise
                samples[i] = signals[i] + errors[i]

        # Return both times and samples, as well as signals and errors
        return samples, signals, errors



__all__ = []


class BaseSignal:
    """BaseSignal class

    Signature for all signal classes.

    """

    def __init__(self):
        raise NotImplementedError

    def sample_next(self, time, samples, errors):
        """Samples next point based on history of samples and errors

        Parameters
        ----------
        time : int
            time
        samples : array-like
            all samples taken so far
        errors : array-like
            all errors sampled so far

        Returns
        -------
        float
            sampled signal for time t

        """
        raise NotImplementedError

    def sample_vectorized(self, time_vector):
        """Samples for all time points in input

        Parameters
        ----------
        time_vector : array like
            all time stamps to be sampled
        
        Returns
        -------
        float
            sampled signal for time t

        """
        raise NotImplementedError



__all__ = []


class BaseNoise:
    """BaseNoise class

    Signature for all noise classes.

    """

    def __init__(self):
        raise NotImplementedError

    def sample_next(self, t, samples, errors):  # We provide t for irregularly sampled timeseries
        """Samples next point based on history of samples and errors

        Parameters
        ----------
        t : int
            time
        samples : array-like
            all samples taken so far
        errors : array-like
            all errors sampled so far

        Returns
        -------
        float
            sampled error for time t

        """
        raise NotImplementedError



import sys
import logging
from setuptools import setup
from setuptools import find_packages
import subprocess


# Manage requirments
with open('requirements.txt') as f:
    requirements = f.read().splitlines()


# Install
setup(name='timesynth',
      version='0.2.3',
      description='Library for creating synthetic time series',
      url='https://github.com/TimeSynth/TimeSynth',
      author='Abhishek Malali, Reinier Maat, Pavlos Protopapas',
      author_email='anon@anon.com',
      license='MIT',
      include_package_data=True,
      packages=find_packages(),
      install_requires=requirements,
      tests_require=['pytest'],
      setup_requires=["pytest-runner"])



import numpy as np
from .base_signal import BaseSignal

__all__ = ['AutoRegressive']


class AutoRegressive(BaseSignal):
    """Sample generator for autoregressive (AR) signals.
    
    Generates time series with an autogressive lag defined by the number of parameters in ar_param.
    NOTE: Only use this for regularly sampled signals
    
    Parameters
    ----------
    ar_param : list (default [None])
        Parameter of the AR(p) process
        [phi_1, phi_2, phi_3, .... phi_p]
    sigma : float (default 1.0)
        Standard deviation of the signal
    start_value : list (default [None])
        Starting value of the AR(p) process
        
    """
    
    def __init__(self, ar_param=[None], sigma=0.5, start_value=[None]):
        self.vectorizable = False
        ar_param.reverse()
        self.ar_param = ar_param
        self.sigma = sigma
        if start_value[0] is None:
            self.start_value = [0 for i in range(len(ar_param))]
        else:
            if len(start_value) != len(ar_param):
                raise ValueError("AR parameters do not match starting value")
            else:
                self.start_value = start_value
        self.previous_value = self.start_value

    def sample_next(self, time, samples, errors):
        """Sample a single time point

        Parameters
        ----------
        time : number
            Time at which a sample was required

        Returns
        -------
        ar_value : float
            sampled signal for time t
        """
        ar_value = [self.previous_value[i] * self.ar_param[i] for i in range(len(self.ar_param))]
        noise = np.random.normal(loc=0.0, scale=self.sigma, size=1)
        ar_value = np.sum(ar_value) + noise
        self.previous_value = self.previous_value[1:]+[ar_value]
        return ar_value
